import logging
import os

import zmq
from helper import config_logger, read_json_from_file_as_dictionary
from zeromqmessagehandler import ZeroMqMessageBus, ZeroMqMessageHandler

from messagehandlerbuilder import ServiceLocatorMessageHandlerBuilder

COMP_NAME = "ServiceLocator"

logger = logging.getLogger(COMP_NAME)


def main():
    config_path = './config.json'
    if "CONFIG_PATH" in os.environ:
        config_path = os.environ["CONFIG_PATH"]

    if os.path.isfile(config_path):
        config = read_json_from_file_as_dictionary(config_path)
    else:
        config = {}

    if "locatorAddress" in config:
        locator_address = config["locatorAddress"]
    else:
        locator_address = "tcp://*:6000"

    log_path = config_path
    if "LOG_PATH" in os.environ:
        log_path = os.environ["LOG_PATH"]
    config_logger(log_path, COMP_NAME, add_error_handler=True)

    context = zmq.Context()

    message_bus = ZeroMqMessageBus(COMP_NAME, context)

    message_handler_builder = ServiceLocatorMessageHandlerBuilder(config, logger)
    message_handler = ZeroMqMessageHandler(context,
                                           COMP_NAME,
                                           message_bus,
                                           message_handler_builder,
                                           service_address=locator_address,
                                           publish_service=False)
    message_handler.handle_messages()


if __name__ == '__main__':
    main()



from messagebus import AckMessage, NakMessage
from messageprocessorutil import UuidMessageProcessor
from .dto import RegisterServiceRequest
from .. import KeepAliveError

from .._ServiceLocator import ServiceLocator


class KeepAliveProcessor(UuidMessageProcessor):

    def __init__(self, service_locator, event_name=None):
        # type: (ServiceLocator, str) -> None
        super(KeepAliveProcessor, self).__init__(event_name)
        self.service_locator = service_locator

    def parse_data(self, data, event_type):
        return data.decode("utf-8")

    def call_business(self, reg_id):
        # type: (str) -> None
        self.service_locator.keep_alive(reg_id)

    def format_response(self, message_bus, message, event, service_name, _):
        message_bus.reply_message(message, AckMessage())

    def format_exception(self, message_bus, message, event, input_obj, exception):
        if isinstance(exception, KeepAliveError):
            message_bus.reply_message(message, NakMessage(exception.message.encode("utf-8")))
        else:
            message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))



import logging
from datetime import datetime, timedelta
from threading import Lock, Thread, Condition
from typing import Dict
from uuid import uuid4

from ._KeepAliveError import KeepAliveError

logger = logging.getLogger("ZeroMqServiceLocator.ServiceLocator")


class RegistrationInfo(object):
    def __init__(self, reg_id, service_name, service_address, registration_expiration):
        self.reg_id = reg_id
        self.service_name = service_name
        self.service_address = service_address
        self.registration_expiration = registration_expiration
        self.last_keep_alive = datetime.now()

    def keep_alive(self):
        self.last_keep_alive = datetime.now()

    def is_expired(self):
        return self.last_keep_alive + timedelta(seconds=self.registration_expiration) < datetime.now()

    def __str__(self):
        return str({
            "reg_id": self.reg_id,
            "service_name": self.service_name,
            "service_address": self.service_address,
            "registration_expiration": self.registration_expiration,
            "last_keep_alive": self.last_keep_alive.isoformat()
        })

    def __repr__(self) -> str:
        return self.__str__()


class ServiceLocator(object):
    def __init__(self, clean_interval=30, registration_expiration=20):
        self.clean_interval = clean_interval
        self.registration_expiration = registration_expiration

        self.lock = Lock()
        self.services = {}  # type: Dict[str, str]
        self.registrations = {}  # type: Dict[str, RegistrationInfo]

        self.clean_thread_running = True
        self.clean_thread_condition = Condition()
        self.clean_thread = Thread(target=self._run_clean_expired_registrations,
                                   name="ServiceLocatorCleanExpiredRegistrations")
        self.clean_thread.daemon = True
        self.clean_thread.start()

    def register_service(self, service_name, address):
        address = address.replace("*", "localhost")
        reg_id = str(uuid4())
        logger.debug("Created registration id: %s", reg_id)
        registration_info = RegistrationInfo(reg_id, service_name, address, self.registration_expiration)

        with self.lock:
            self.services[service_name] = address
            self.registrations[reg_id] = registration_info

        logger.debug("Current registrations: %s", self.registrations)
        logger.debug("Current services: %s", self.services)

        return reg_id

    def locate_service(self, service_name):
        with self.lock:
            return self.services.get(service_name, None)

    def keep_alive(self, reg_id):
        try:
            with self.lock:
                registration_info = self.registrations[reg_id]
                current_address = self.services[registration_info.service_name]

            if current_address != registration_info.service_address:
                raise KeepAliveError("The address this registration refers to is not the current address. "
                                     "Current Address: {}".format(current_address))
            registration_info.keep_alive()

            logger.debug("Keep alive processed: %s", registration_info)
            logger.debug("Current registrations: %s", self.registrations)
            logger.debug("Current services: %s", self.services)
        except KeyError as k:
            if k.args[0] == reg_id:
                raise KeepAliveError("The registration id {} is not registered anymore".format(reg_id))

            logger.debug("Current registrations: %s", self.registrations)
            logger.debug("Current services: %s", self.services)
            raise KeepAliveError("The registration is active but the service is not registered.".format(reg_id))

    def dispose(self):
        self.clean_thread_running = False
        self.clean_thread_condition.notify_all()
        self.clean_thread.join()

    def _run_clean_expired_registrations(self):
        while self.clean_thread_condition:
            try:
                self._clean_expired_registrations()
                with self.clean_thread_condition:
                    self.clean_thread_condition.wait(self.clean_interval)
            except:
                logger.exception("Error cleaning expired registrations")

    def _clean_expired_registrations(self):
        logger.debug("Running clean")
        logger.debug("Current registrations: %s", self.registrations)
        logger.debug("Current services: %s", self.services)
        with self.lock:
            to_remove = []
            services_to_keep = {}
            for reg_id in self.registrations.keys():
                registration_info = self.registrations[reg_id]
                if registration_info.is_expired():
                    to_remove.append(registration_info)
                else:
                    services_to_keep[registration_info.service_name] = registration_info.service_name

            for registration_info in to_remove:
                logger.debug("Removing expired registration: %s", registration_info)
                del self.registrations[registration_info.reg_id]
                if registration_info.service_name not in services_to_keep:
                    del self.services[registration_info.service_name]
        logger.debug("Cleaned services: %s", self.services)
        logger.debug("Cleaned registrations: %s", self.registrations)



from logging import Logger

from messagehandler import MessageHandlerBuilder
from messageprocessor import MessageProcessorMessageHandler, DefaultMessageProcessorExecutorFactory
from messageprocessorutil import LoggingProcessorCallback
from zeromqservicelocator import \
    ServiceLocator, \
    TK_SERVICE_LOCATOR_REGISTER_SERVICE, \
    TK_SERVICE_LOCATOR_LOOKUP_SERVICE, \
    TK_SERVICE_LOCATOR_KEEP_ALIVE
from zeromqservicelocator.processor import RegisterServiceProcessor, LocateServiceProcessor
from zeromqservicelocator.processor import KeepAliveProcessor


class ServiceLocatorMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(self, config, logger):
        # type: (dict, Logger) -> None
        self.config = config
        self.logger = logger

        self.service_locator = None  # type: ServiceLocator

    def build_singletons(self):
        self.service_locator = ServiceLocator()

    def build_message_handler(self, message, event):
        handled_tokens = {
            TK_SERVICE_LOCATOR_REGISTER_SERVICE: RegisterServiceProcessor(self.service_locator),
            TK_SERVICE_LOCATOR_LOOKUP_SERVICE: LocateServiceProcessor(self.service_locator),
            TK_SERVICE_LOCATOR_KEEP_ALIVE: KeepAliveProcessor(self.service_locator),
        }
        return MessageProcessorMessageHandler(
            None,
            handled_tokens,
            executor_factory=DefaultMessageProcessorExecutorFactory([LoggingProcessorCallback(self.logger)]))

    def destroy_message_handler(self, message_handler):
        pass

    def destroy_singletons(self):
        pass



from messagebus import AckMessage, NakMessage
from messageprocessorutil import UuidMessageProcessor
from .dto import RegisterServiceRequest

from .._ServiceLocator import ServiceLocator


class LocateServiceProcessor(UuidMessageProcessor):

    def __init__(self, service_locator, event_name=None):
        # type: (ServiceLocator, str) -> None
        super(LocateServiceProcessor, self).__init__(event_name)
        self.service_locator = service_locator

    def parse_data(self, data, event_type):
        service_name = data.decode("utf-8")
        return service_name

    def call_business(self, service_name):
        # type: (str) -> None
        return self.service_locator.locate_service(service_name)

    def format_response(self, message_bus, message, event, service_name, service_address):
        if service_address is not None:
            message_bus.reply_message(message, AckMessage(data=service_address.encode("utf-8")))
        else:
            message_bus.reply_message(message, NakMessage("Service {} not found".format(service_name).encode("utf-8")))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))



from messagebus import AckMessage, NakMessage
from messageprocessorutil import UuidMessageProcessor
from .dto import RegisterServiceRequest

from .._ServiceLocator import ServiceLocator


class RegisterServiceProcessor(UuidMessageProcessor):

    def __init__(self, service_locator, event_name=None):
        # type: (ServiceLocator, str) -> None
        super(RegisterServiceProcessor, self).__init__(event_name)
        self.service_locator = service_locator

    def parse_data(self, data, event_type):
        parts = data.split(b"\0")
        if len(parts) < 2:
            raise Exception("Invalid register message")
        service_name = parts[0].decode("utf-8")
        service_address = parts[1].decode("utf-8")
        return RegisterServiceRequest(service_name, service_address)

    def call_business(self, request):
        # type: (RegisterServiceRequest) -> None
        return self.service_locator.register_service(request.service_name, request.address)

    def format_response(self, message_bus, message, event, input_obj, reg_id):
        message_bus.reply_message(message, AckMessage(reg_id.encode("utf-8")))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))



from setuptools import setup, find_packages

setup(name="ZeroMqServiceLocator",
      version="2.0.0",
      packages=find_packages(where="."),
      py_modules=["main"],
      install_requires=['ZeroMqMessageHandler==2.0.0', 'EdpHelper==0.15', 'MessageProcessor==3.0.2'],
      url="http://www.e-deploy.com.br",
      author="Eduardo Almeida",
      author_email="eduardo.almeida@e-deploy.com.br")



from ._ServiceLocator import ServiceLocator
from ._KeepAliveError import KeepAliveError
from ._ServiceLocatorTokens import \
    TK_SERVICE_LOCATOR_LOOKUP_SERVICE, \
    TK_SERVICE_LOCATOR_REGISTER_SERVICE, \
    TK_SERVICE_LOCATOR_KEEP_ALIVE
